import pickle

from xgboost import XGBRegressor, XGBClassifier

from sklearn.model_selection import train_test_split
from sklearn.metrics import mean_squared_error, accuracy_score
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
from sklearn.neighbors import KNeighborsClassifier, KNeighborsRegressor
from sklearn.ensemble import GradientBoostingClassifier, GradientBoostingRegressor, AdaBoostClassifier, AdaBoostRegressor, RandomForestClassifier, RandomForestRegressor
from sklearn import svm
from sklearn.linear_model import LogisticRegression, LinearRegression, Ridge, Lasso, SGDClassifier, SGDRegressor



# ----------------------------------------------------------------- DT
def DecisionTree(task, data, split=0.3, max_depths=None, seed=42):
    """
    Simplified Decision Tree
    
    Parameters
    ----------
    task : str 
        String describing if the task is Classification or Regression
    data : dict
        Dictionary containing features and values for given features
    split : float
        Train/Test Split for the data inside dict
    max_depths : int
        Maximum Depth the Decision Tree can go to
    seed : int
        Value that controls shuffling of data
    

    Returns
    -------
    model
        The actual Decision Tree Model fitted to the training data
    
    """ 
    
    task = task.lower()
    X = data["X"]
    y = data["y"]

    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=split, random_state=seed)
    
    if task == "c" or task == "classify" or task == "classification":
        model = DecisionTreeClassifier(max_depth=max_depths, random_state=seed)
        model.fit(X_train, y_train)
        
        train_preds = model.score(X_train, y_train)
        print("Decision Tree Training Accuracy: " + str(train_preds*100) + "%")
        print("Decision Tree Testing Accuracy:  " + str(model.score(X_test, y_test) * 100) + "%") 

        return model

    elif task == "r" or task == "reg" or task == "regression":
        model = DecisionTreeRegressor(max_depth=max_depths, random_state=seed)
        model.fit(X_train, y_train)

        train_preds = model.predict(X_train)
        train_rmse = mean_squared_error(y_train, train_preds, squared=False)
        print("Decision Tree Training RMSE: " + str(train_rmse) + "")

        test_preds = model.predict(X_test)
        test_rmse = mean_squared_error(y_test, test_preds, squared=False)
        print("Decision Tree Testing RMSE: " + str(test_rmse) + "")
        
        return model     
    else:
        raise NameError('Task should be classification or regression')


# ----------------------------------------------------------------- KNN

def KNN(task, data, split=0.3, neighbors=5, weights="uniform", seed=42): 
    """
    Simple Implementation of a KNearestNeighbors

    Parameters
    ----------
    task : str 
        String describing if the task is Classification or Regression
    data : dict
        Dictionary containing features and values for given features
    split : float
        Train/Test Split for the data inside dict
    neighbors : int
        Number of Neighbors for the KNN Algorithm
    weights : str
        Weight Function used in calculating final KNN Prediction
    seed : int
        Value that controls shuffling of data

    Returns
    -------
    model
        The actual KNN Model fitted to the training data
    
    """ 
    
    task = task.lower()
    X = data["X"]
    y = data["y"]

    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=split, random_state=seed)

    if task == "c" or task == "classify" or task == "classification":
        model = KNeighborsClassifier(n_neighbors=neighbors, weights=weights)
        model.fit(X_train, y_train)
        
        train_preds = model.score(X_train, y_train)
        print("KNN Training Accuracy: " + str(train_preds*100) + "%")
        print("KNN Testing Accuracy: " + str(model.score(X_test, y_test) * 100) + "%")
    
        return model

    elif task == "r" or task == "reg" or task == "regression":
        model = KNeighborsRegressor(n_neighbors=neighbors, weights=weights)
        model.fit(X_train, y_train)

        train_preds = model.predict(X_train)
        train_rmse = mean_squared_error(y_train, train_preds, squared=False)
        print("KNN Training RMSE: " + str(train_rmse))

        test_preds = model.predict(X_test)
        test_rmse = mean_squared_error(y_test, test_preds, squared=False)
        print("KNN Testing RMSE: " + str(test_rmse))

    else:
        raise('Task should be classification or regression')

# ----------------------------------------------------------------- Random Forests
def RandomForest(task, data, split=0.3, n_estimators=100, max_depths=None, seed=42):
    """
    Easy to Use Random Forest Algorithm

    Parameters
    ----------
    task : str 
        String describing if the task is Classification or Regression
    data : dict
        Dictionary containing features and values for given features
    split : float
        Train/Test Split for the data inside dict
    n_estimators : int
        Number of Trees within the Forest
    max_depths : int
        Maximum Depth each Tree can go to
    seed : int
        Value that controls shuffling of data


    Returns
    -------
    model
        The Actual Random Forest Model fitted to the training data
    
    """

    task = task.lower()
    X = data["X"]
    y = data["y"]

    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=split, random_state=seed)
    
    if task == "c" or task == "classify" or task == "classification":
        model = RandomForestClassifier(n_estimators=n_estimators,max_depth=max_depths, random_state=seed)
        model.fit(X_train, y_train)

        train_preds = model.score(X_train, y_train)
        print("RandomForest Training Accuracy: " + str(train_preds*100) + "%")
        print("RandomForest Testing Accuracy:  " + str(model.score(X_test, y_test) * 100) + "%")

        return model

    elif task == "r" or task == "reg" or task == "regression":
        model = RandomForestRegressor(n_estimators=n_estimators, max_depth=max_depths, random_state=seed)
        model.fit(X_train, y_train)

        train_preds = model.predict(X_train)
        train_rmse = mean_squared_error(y_train, train_preds, squared=False)
        print("RandomForest Training RMSE: " + str(train_rmse) + "")

        test_preds = model.predict(X_test)
        test_rmse = mean_squared_error(y_test, test_preds, squared=False)
        print("RandomForest Testing RMSE: " + str(test_rmse) + "")
        
        return model  

    else:
        raise NameError('Task should be Regression or Classification')

# ----------------------------------------------------------------- Boosting
def Boosting(task, data, split=0.3, algo="xgb", n_estimators=75, lr=0.5, seed=42):
    """
    Simplified Boosting Algorithm

    Parameters
    ----------
    task : str 
        String describing if the task is Classification or Regression
    data : dict
        Dictionary containing features and values for given features
    split : float
        Train/Test Split for the data inside dict
    algo : str
        Specific Boosting Algorithm to be used[XGBoost, Gradient Boosting]
    n_estimators : int
        Number of Trees within the Forest
    lr : float
        The amount the contribution of each tree(n_estimators) is decreased by
    seed : int
        Value that controls shuffling of data

    Returns
    -------
    model
        The Actual Boosting Model fitted to the training data

    """ 
    
    algo = algo.lower()
    task = task.lower()
    X = data["X"]
    y = data["y"]

    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=split, random_state=seed)

    if algo == "xgb" or task == "xgboost":
        if task == "c" or task == "classify" or task == "classification":
            model = XGBClassifier(n_estimators=n_estimators, learning_rate=lr, random_state=seed)
            model.fit(X_train, y_train)

            train_preds = model.score(X_train, y_train)
            print("XGBoost Training Accuracy: " + str(train_preds*100) + "%")
            print("XGBoost Testing Accuracy: " + str(model.score(X_test, y_test)*100) + "%")

            return model 

        elif task == "r" or task == "reg" or task == "regression":
            model = XGBRegressor(n_estimators=n_estimators, learning_rate=lr, random_state=seed)
            model.fit(X_train, y_train)
            
            train_preds = model.predict(X_train)
            train_rmse = mean_squared_error(y_train, train_preds, squared=False)
            print("XGBoost Training RMSE: " + str(train_rmse))

            test_preds = model.predict(X_test)
            test_rmse = mean_squared_error(y_test, test_preds, squared=False)
            print("XGBoost Testing RMSE: " + str(test_rmse))

            return model

        else: 
            raise NameError('Task should be Regression or Classification')
   
    elif algo == "gb" or algo == "gradient" or algo == "gradientboosting":
        if task == "c" or task == "classify" or task == "classification":
            model = GradientBoostingClassifier(n_estimators=n_estimators, learning_rate=lr, random_state=seed)
            model.fit(X_train, y_train)

            train_preds = model.score(X_train, y_train)
            print("Gradient Boosting Training Accuracy: " + str(train_preds*100) + "%")
            print("Gradient Boosting Testing Accuracy: " + str(model.score(X_test, y_test)*100) + "%")

            return model 

        elif task == "r" or task == "reg" or task == "regression":
            model = GradientBoostingRegressor(n_estimators=n_estimators, learning_rate=lr, random_state=seed)
            model.fit(X_train, y_train)
            
            train_preds = model.predict(X_train)
            train_rmse = mean_squared_error(y_train, train_preds, squared=False)
            print("Gradient Boosting Training RMSE: " + str(train_rmse))

            test_preds = model.predict(X_test)
            test_rmse = mean_squared_error(y_test, test_preds, squared=False)
            print("Gradient Boosting Testing RMSE: " + str(test_rmse))

            return model

        else:
            raise NameError('Task should be Regression or Classification')

    elif algo == "ada" or algo == "adaboost":
        if task == "c" or task == "classify" or task == "classification":
            model = AdaBoostClassifier(n_estimators=n_estimators, learning_rate=lr, random_state=42)
            model.fit(X_train, y_train)

            train_preds = model.score(X_train, y_train)
            print("AdaBoost Training Accuracy: " + str(train_preds*100) + "%")
            print("AdaBoost Testing Accuracy: " + str(model.score(X_test, y_test)*100) + "%")

            return model 

        elif task == "r" or task == "reg" or task == "regression":
            model = AdaBoostRegressor(n_estimators=n_estimators, learning_rate=lr, random_state=42)
            model.fit(X_train, y_train)
            
            train_preds = model.predict(X_train)
            train_rmse = mean_squared_error(y_train, train_preds, squared=False)
            print("AdaBoost Training RMSE: " + str(train_rmse))

            test_preds = model.predict(X_test)
            test_rmse = mean_squared_error(y_test, test_preds, squared=False)
            print("AdaBoost Testing RMSE: " + str(test_rmse))

            return model

        else:
            raise NameError('Task should be Regression or Classification')

    else:
        raise NameError('Algorithm should be AdaBoost, Gradient Boosting or XGBoost')



# -------------------------------------------------------------------------- SGD
def SGD(task, data, split=0.3, alpha=0.0001, seed=42):
    """
    Stochastic Gradient Descent 

    Parameters
    ----------
    task : str 
        String describing if the task is Classification or Regression
    data : dict
        Dictionary containing features and values for given features
    split : float
        Train/Test Split for the data inside dict
    alpha : float
        Constant that multiplies the regularization term
    seed : int
        Value that controls shuffling of data

    Returns
    -------
    model
        SGD Model Fitted to Training Data

    """ 
    
    task = task.lower()
    X = data["X"]
    y = data["y"]

    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=split, random_state=seed)
    
    if task == "c" or task == "classify" or task == "classification":
        model = SGDClassifier(alpha=alpha, random_state=seed)
        model.fit(X_train, y_train)

        train_preds = model.score(X_train, y_train)
        print("SGD Training Accuracy: " + str(train_preds*100) + "%")
        print("SGD Testing Accuracy: " + str(model.score(X_test, y_test) * 100) + "%")

        return model

    elif task == "r" or task == "reg" or task == "regression":
        model = SGDRegressor(alpha=alpha, random_state=seed)
        model.fit(X_train, y_train)

        train_preds = model.predict(X_train)
        train_rmse = mean_squared_error(y_train, train_preds, square=False)
        print("SGD Training RMSE: " + str(train_rmse))

        test_preds = model.predict(X_test)
        test_rmse = mean_squared_error(y_test, test_preds, squared=False)
        print("SGD Testing RMSE: " + str(test_rmse))

        return model
  
    else:
        raise NameError('Task should be either regression or classification')

# -------------------------------------------------------------------------- SVMs
def SVM(task, data, split=0.3, C=1, seed=42):
    """
    Easy to Use Support Vector Machine

    Parameters
    ----------
    task : str 
        String describing if the task is Classification or Regression
    data : dict
        Dictionary containing features and values for given features
    split : float
        Train/Test Split for the data inside dict
    C : float
        Regularization Parameter
    seed : int
        Value that controls shuffling of data


    Returns
    -------
    model 
        The Actual SVM Model fitted to the Training Data

    """


    task = task.lower()
    X = data["X"]
    y = data["y"]

    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=split, random_state=seed)

    if task == "c" or task == "classify" or task == "classification":
        model = svm.SVC(C=C)
        model.fit(X_train, y_train)
        
        train_preds = model.score(X_train, y_train)
        print("Support Vector Machines Training Accuracy: " + str(train_preds*100) + "%")
        print("Support Vector Machines Testing Accuracy: " + str(model.score(X_test, y_test)*100) + "%")

        return model
    elif task == "r" or task == "reg" or task == "regression":
        model = svm.SVR(C=C)
        model.fit(X_train, y_train)

        train_preds = model.predict(X_train)
        train_rmse = mean_squared_error(y_train, train_preds, squared=False)
        print("Support Vector Machines Training RMSE: " + str(train_rmse))

        test_preds = model.predict(X_test)
        test_rmse = mean_squared_error(y_test, test_preds, squared=False)
        print("Support Vector Machines Testing RMSE: " + str(test_rmse))

        return model

    else:
        raise NameError('task should be either regression or classification')


# ------------------------------------------------------------------------ Logistic Regression
def LogisticReg(data, split=0.3, solver="lbfgs", seed=42):
    """
    Logistic Regression Model

    Parameters
    ----------
    data : dict
        Dictionary containing features and values for given features
    split : float
        Train/Test Split for the data inside dict
    solver : str
        Algorithm used for Optimization problem for Log. Reg
        Other Solvers:
            lbfgs
            newton-cg
            lib-linear
            sag
            saga
    seed : int
        Value that controls shuffling of data
    
    Returns
    -------
    model
        Logistic Regression Model fitted to the training data

    """

    X = data["X"]
    y = data["y"]

    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=split, random_state=seed)

    model = LogisticRegression(solver=solver, random_state=42)
    model.fit(X_train, y_train)
    train_preds = model.score(X_train, y_train)
    print("Logistic Regression Training Accuracy: " + str(train_preds * 100) + "%")
    test_preds = model.score(X_test, y_test)
    print("Logistic Regression Testing Accuracy: " + str(test_preds * 100) + "%")
    
    return model 


# ------------------------------------------------------------------------ Regression Models
def Regression(data, split=0.3, task="linear", seed=42):
    """
    Regression Model

    Parameters
    ----------
    data : dict
        Dictionary containing features and values for given features
    split : float
        Train/Test Split for the data inside dict
    task : str
        Variable so you can choose between different types of Regression
        Options:
            Linear
            Lasso
            Ridge
    seed : int
        Value that controls shuffling of data

    Returns
    -------
    model
        Regression Model fitted to training data

    """ 
    
    task = task.lower()
    X = data["X"]
    y = data["y"]

    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=split, random_state=seed)

    if task == "linear" or task == "li":
        model = LinearRegression()
        model.fit(X_train, y_train)

        train_preds = model.predict(X_train)
        train_rmse = mean_squared_error(y_train, train_preds, squared=False)
        print("Linear Regression Training RMSE: " + str(train_rmse))

        test_preds = model.predict(X_test)
        test_rmse = mean_squared_error(y_test,test_preds,squared=False)
        print("Linear Regression Testing RMSE: " + str(test_rmse))

        return model
    elif task == "ridge" or task == "r":
        model = Ridge()
        model.fit(X_train, y_train)

        train_preds = model.predict(X_train)
        train_rmse = mean_squared_error(y_train, train_preds, squared=False)
        print("Ridge Regression Training RMSE: " + str(train_rmse))

        test_preds = model.predict(X_test)
        test_rmse = mean_squared_error(y_test,test_preds,squared=False)
        print("Ridge Regression Testing RMSE: " + str(test_rmse))

        return model

    elif task == "lasso" or task == "la":
        model = Lasso()
        model.fit(X_train, y_train)

        train_preds = model.predict(X_train)
        train_rmse = mean_squared_error(y_train, train_preds, squared=False)
        print("Lasso Regression Training RMSE: " + str(train_rmse))

        test_preds = model.predict(X_test)
        test_rmse = mean_squared_error(y_test,test_preds,squared=False)
        print("Lasso Regression Testing RMSE: " + str(test_rmse))

        return model
    else:
        raise NameError('Specify a correct regression algorithm')

# ------------------------------------------------------------------------- Mode Hel
def save_model(model):
    with open('model.pkl', 'wb') as fid:
        pickle.dump(model, fid)
        print("Saved to disk!")


def load_model(path):
    with open(path, 'rb') as fid:
        model = pickle.load(fid)
    print("Loaded!")
    return model



import os

import pandas as pd
import numpy as np

import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D as ax
import plotly.express as px
import plotly.graph_objects as go
from pandas.plotting import scatter_matrix
import seaborn as sns

from tensorflow.keras.datasets import fashion_mnist, mnist
from tensorflow.keras.preprocessing.image import ImageDataGenerator

from sklearn import datasets
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import MinMaxScaler, StandardScaler

import warnings

warnings.filterwarnings("ignore")

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'


def plot_subplots(df, cols, num=5):
    """
    Helper method for data visualization

    Parameters
    ----------
    df : pandas DataFrame
        Training data with column names
    cols : list
        List of column names
    num : int
        Number of features to include in subplots

    Returns
    -------
    No return values, just plots the matrix

    """
    scatter_matrix(df[cols[: num]], figsize=(12, 8))
    plt.savefig("features.png")


def correlations(df, cols):
    """
    Helper method for data visualization

    Parameters
    ----------
    df : pandas DataFrame
        Training data with column names
    cols : list
        List of column names

    Returns
    -------
    imgs : matplotlib Axes
        Image containing correlation diagram
    Figure is also saved locally as "correlation.png"

    """
    plt.figure(figsize=(16, 12))
    imgs = sns.heatmap(df[cols].corr(), cmap="YlGnBu", annot=True, fmt='.2f', vmin=0)
    plt.savefig("correlation.png")
    return imgs


def load_dataset(name, num_features=5, random_state=42, flatten=False, show_corr_matrix=True, show_subplots=True):
    """
    Loading Common Datasets
    
    Parameters
    ----------
    name : str 
        Name of dataset ('mnist', 'fmnist', 'iris', 'breast_cancer', 'diabetes', 'housing')
    num_features : int
        Number of features to view in subplots, if None then num_features includes all features
    random_state : int
        Specify random state
    flatten : bool
        Returns image with shape (-1, 28, 28) if True, else returns image with shape (-1, 784)
    show_corr_matrix : bool
        Whether to show correlation matrix
    show_subplots : bool
        Whether to show subplots comparing the num_features
    

    Returns
    -------
    (X_train, y_train)
        Training data in the form of numpy arrays
    (X_test, y_test)
        Testing data (20% of total data)
    col_names
        Feature names (None if dataset is MNIST or Fashion MNIST)
    
    """ 
    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'
    if name == 'mnist':
        (X_train, y_train), (X_test, y_test) = mnist.load_data()
        if flatten:
            X_train = X_train.reshape(-1, 784)
            X_test = X_test.reshape(-1, 784)
            X_train = X_train / 255.0
            X_test = X_test / 255.0
        return (X_train, y_train), (X_test, y_test), None

    elif name == 'fashion_mnist' or name == 'fmnist':
        (X_train, y_train), (X_test, y_test) = fashion_mnist.load_data()

        if flatten:
            X_train = X_train.reshape(-1, 784)
            X_test = X_test.reshape(-1, 784)
        return (X_train, y_train), (X_test, y_test), None

    elif name == 'iris':
        data = datasets.load_iris()
        X = data.data
        y = data.target
        col_names = data.feature_names
        df = pd.DataFrame(X, columns=col_names)
        num = num_features
        if num_features == None:
            num = len(col_names)

        if show_subplots:
            plot_subplots(df, col_names, num)
        if show_corr_matrix:
            corr_img = correlations(df, col_names)

        X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=42)

    elif name == 'cancer' or name == 'breast_cancer':
        data = datasets.load_breast_cancer()
        X = data.data
        y = data.target
        col_names = data.feature_names
        df = pd.DataFrame(X, columns=col_names)
        num = num_features
        if num_features == None:
            num = len(col_names)
        if show_subplots:
            plot_subplots(df, col_names, num)
        if show_corr_matrix:
            corr_img = correlations(df, col_names)
        X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=42)

    elif name == 'diabetes':
        data = datasets.load_diabetes()
        X = data.data
        y = data.target
        col_names = data.feature_names
        df = pd.DataFrame(X, columns=col_names)
        num = num_features
        if num_features == None:
            num = len(col_names)

        if show_subplots:
            plot_subplots(df, col_names, num)
        if show_corr_matrix:
            corr_img = correlations(df, col_names)
        X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=42)

    elif name == 'house' or name == 'california_housing' or name == 'housing':
        data = datasets.fetch_california_housing()
        X = data.data
        y = data.target
        col_names = data.feature_names
        df = pd.DataFrame(X, columns=col_names)
        num = num_features
        if num_features == None:
            num = len(col_names)
        if show_subplots:
            plot_subplots(df, col_names, num)
        if show_corr_matrix:
            corr_img = correlations(df, col_names)
        X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=42)
        if show_corr_matrix:
            plt.title("Correlation Matrix")
    
    else:
        raise NameError("Please provide an appropriate value for 'name'")
    
    return (X_train, y_train), (X_test, y_test), col_names


def load_custom_data(path, predicted_variable, num_features=5, random_state=42, show_corr_matrix=True,
                     show_subplots=True, id=False):

    """ 
    Makes it easy to load custom datasets[mainly CSV's]

    Parameters
    ----------
    path : str
        Path to the dataset
    predicted_variable : str
        The variable you are trying to get the output for
    num_features : int
        Number of features that you want to see on the Scatter Matrix
    random_state : int
        Seed for the validation set(to be added)
    show_corr_matrix : bool
        If you want to see the correlation matrx or not
    show_subplots : bool
        If you want to see subplots or not

    Returns
    -------
    data : dict
        Contains X, and Y data
    col_names : List
        Contains a list of the column names

    """    


    df = pd.read_csv(path)
    X = df.drop(predicted_variable, axis=1)
    if id:
        X = X.drop("id", axis=1)
    y = df[predicted_variable]
    col_names = X.columns
    num = num_features
    if num_features == None:
        num = len(col_names)

    if show_subplots:
        plot_subplots(df, col_names, num)

    if show_corr_matrix:
        corr_img = correlations(df, col_names)

    data = {
        "X": X,
        "y": y,
    }



    return data, col_names


def kaggle_submission_csv(model, test_path, X, y, start=True):
    """
    Makes it easy to submit to kaggle.com

    Parameters
    ----------
    model : (ML_MODEL) 
        The name of the machine learning model you trained on the dataset
    test_path : str
        Path to the testing data
    X : str
        Name of the column containing all of the numbers(to identify testing data)
    y : str
        Name of the column containing the output of the testing data(corresponding to X)
    
    Returns
    -------
    Nothing, but converts it to a csv

    """

    test = pd.read_csv(test_path)
    print(test.shape)
    columns = test.columns
    predictions = (model.predict(test[columns]))
    if start:
        submit = pd.DataFrame({X: range(0, predictions.size), y: predictions})
        submit.to_csv("submission.csv", index=False)
    else:
        submit = pd.DataFrame({X: range(1, predictions.size + 1), y: predictions})
        submit.to_csv("submission.csv", index=False)
    print("Conversion Successful")


def visualize_data(X, y, column_index=0, task="scatter", color=None):
    """
    Visualize Data in a various forms

    Parameters
    ----------
    X : np.array
        The x_axis you want(or first comparator for barplot)
    y : np.array
        The y_axis you want(or second compartor for barplot)
    column_index : int
        The specific column from the X training data you are analyzing
    task : str
        Specific Task you are trying to accomplish[Line, Scatter, Bar and Histogram]
    color : col_name
        Distinguishing different values(countries, flowers, etc.)
    test_path : str
        Path to the Testing Data

    Returns
    -------
    Nothing, just shows the Graph!

    """

    flag = False
    if task == "line":
        fig = px.line(x=X[:, column_index], y=y, color=color)
        flag = True
    elif task == "scatter":
        fig = px.scatter(x=X[:, column_index], y=y)
        flag = True
    elif task == "bar":
        fig = px.bar(x=X[:, column_index], y=y)
        flag = True
    elif task == "hist":
        fig = px.histogram(x=X[:, column_index])
        flag = True
    elif task == "compBP":
        sns.barplot(x=X, y=y)
        plt.savefig("BarPlot.png")
    else:
        raise NameError("Please provide an appropriate value for 'task'")

    if flag:
        fig.show()


def plot_3d(x, y, z):
    """
    Plot Points in 3D

    Parameters
    ----------
    x : np.array
        Numpy Array with shape (n,)
    y : np.array
        Numpy Array with shape (n,)
    z : np.array
        Numpy Array with shape (n,)

    Returns
    -------
        Returns 3D PlotLy Scatter Plot
    """
    
    fig = px.scatter_3d(x=x, y=y, z=z)
    fig.show()


def regression_data(X_min=0, X_max=2, n_samples=100, n_features=1, noise='high'):
    """
    Generate Regression Data 

    Parameters
    ----------
    X_min : [int or List]
        if list, length of X_min should be n_features, specifies bounds for each feature
        if value specified is int, then X_min is same for all features
    X_max : [int or List]
        if list, length of X_max should be n_features, specifies bounds for each feature
        if value specified is int, then X_max is same for all features
    n_samples : int
        Number of points to generate
    n_features : int
        Number of Features
    noise : str
        3 Possible Levels['high', 'medium', 'low']
        Dictates the level of noise
    
    Returns
    -------
    (X_train, y_train) : np.array
        Numpy array with training data
    (X_test, y_test) : np.array
        Numpy array with test data
    if n_features is 1: returns scatter plot with data (PlotLy)
    if n_features is 2: returns 3d scatter plot with data (PlotLy)
    if n_features > 2: return 2d plot comparing first feature with target value
    """

    if type(X_max) == 'list':
        data_range = np.array(X_max) - np.array(X_min)
    else:
        data_range = np.ones(n_features) * (X_max - X_min)
        X_min = np.ones(n_features) * (X_min)

    df = {}
    df['X1'] = (data_range[0]) * np.random.rand(n_samples) + X_min[0]

    coef = 20 * np.random.rand() - 10
    bias = (100) * np.random.rand() - 100
    if noise == 'high':
        y = bias + coef * df['X1'] + 1 * coef * data_range[0] * np.random.rand(n_samples) - X_min[0]

    elif noise == 'medium':
        y = bias + coef * df['X1'] + 0.7 * coef * data_range[0] * np.random.rand(n_samples) - X_min[0]

    elif noise == 'low':
        y = bias + coef * df['X1'] + 0.5 * coef * data_range[0] * np.random.rand(n_samples) - X_min[0]

    if n_features > 1:
        for i in range(n_features - 1):
            coef = 20 * np.random.rand() - 10
            index = str(i + 2)
            val = f"X{index}"
            df[val] = ((data_range[i + 1]) * np.random.rand(n_samples) + X_min[i + 1])
            y += coef * df[f'X{i + 2}']
    df['y'] = y
    df = pd.DataFrame(df)

    if n_features == 2:
        plot_3d(df['X1'], df['X2'], df['y'])

    if n_features != 2:
        fig = px.scatter(x=df['X1'], y=y)
        fig.show()
    X_train, X_test, y_train, y_test = train_test_split(np.array(df.drop('y', axis=1, inplace=False)),
                                                        np.array(df['y']), test_size=0.2)
    return (X_train, y_train), (X_test, y_test)


def classification_data(n_clusters=2, n_samples=100):
    """
    Generate Classification Data

    Parameters
    ----------
    n_clusters : int 
        Numer of clusters to generate
    n_samples : int 
        Number of points to generate in each cluster
    
    Returns
    -------
    (X_train, y_train) : np.array
        Numpy Array with Training Data
    (X_test, y_test) : np.array
        Numpy array with test data
    
    Can also display 2D scatter plot with clusters
    """

    distance = 50 * (np.log(n_clusters))
    data = {}
    data['x'] = []
    data['y'] = []
    data['class'] = []
    for i in range(n_clusters):
        center1 = (100 * ((n_clusters / 2) + 2 * n_clusters) * np.random.rand(),
                   (100 * ((n_clusters / 2) + 2 * n_clusters) * np.random.rand()))
        x1 = np.random.uniform(center1[0], center1[0] + distance, size=(n_samples,))

        for x in x1:
            data['x'].append(x)
        y1 = np.random.normal(center1[1], distance, size=(n_samples,))

        for y in y1:
            data['y'].append(y)

        for label in np.ones(n_samples) * i:
            data['class'].append(label)

    data = pd.DataFrame(data)

    data_2 = data.sample(frac=1).reset_index(drop=True)
    arr = np.array(data_2)
    data['class'] = [str(int(label)) for label in data['class']]
    fig = px.scatter(data, x='x', y='y', color='class')
    fig.show()
    X_train, X_test, y_train, y_test = train_test_split(np.array(data.drop(['class'], axis=1, inplace=False)),
                                                        np.array(data['class']), test_size=0.2)
    return (X_train, y_train), (X_test, y_test)


def read_images_from_dataframe(df, IMAGE_DIR, file_col='files', class_col='class', class_mode='raw',
                               target_size=(224, 224), batch_size=32, shuffle=True, validation_split=0.2):
    """
    Loading Common Datasets

    Parameters
    ----------
    df : pandas DataFrame
        DataFrame containing labels and filenames
    IMAGE_DIR : str
        The path to the directory containing images
    file_col : str
        Name of column containing image files from df
    class_col : str
        Name of column containing ground truth values from df
    class_mode : str
        One of the following options: ["binary", "categorical", "input", "multi_output", "raw", sparse", None]
    target_size : tuple
        Input dimension of network. All images will be reshaped to this size.
    batch_size : int
        Number of images in each training batch
    shuffle : bool
        Whether to shuffle images presented in the DataFrame
    validation_split : float
        The fraction of images to include for validation


    Returns
    -------
    train_generator
        Generator that returns batches of training data
    test_generator
        Generator that returns data for testing/validation

    """
    img_datagen = ImageDataGenerator(rescale=1. / 255,
                                     validation_split=validation_split)

    train_generator = img_datagen.flow_from_dataframe(df,
                                                      directory=IMAGE_DIR,
                                                      x_col=file_col,
                                                      y_col=class_col,
                                                      target_size=target_size,
                                                      batch_size=batch_size,
                                                      shuffle=shuffle,
                                                      class_mode=class_mode,
                                                      subset='training')

    test_generator = img_datagen.flow_from_dataframe(df,
                                                     directory=IMAGE_DIR,
                                                     x_col=file_col,
                                                     y_col=class_col,
                                                     target_size=target_size,
                                                     batch_size=batch_size,
                                                     shuffle=shuffle,
                                                     class_mode=class_mode,
                                                     subset='validation')

    return train_generator, test_generator


def read_images_from_directory(IMAGE_DIR, target_size=(100, 100), class_mode='categorical',
                               batch_size=16, shuffle=True, validation_split=0.2):
    """
    Loading Common Datasets

    Parameters
    ----------
    IMAGE_DIR : str
        The path to the directory containing images
    target_size : tuple
        Input dimension of network. All images will be reshaped to this size.
    class_mode : str
        One of the following options: ["binary", "categorical", "input", "multi_output", "raw", sparse", None]
    batch_size : int
        Number of images in each training batch
    shuffle : bool
        Whether to shuffle images presented in the DataFrame
    validation_split : float
        The fraction of images to include for validation


    Returns
    -------
    train_generator
        Generator that returns batches of training data
    test_generator
        Generator that returns data for testing/validation

    """
    img_datagen = ImageDataGenerator(rescale=1. / 255,
                                     validation_split=validation_split)

    train_generator = img_datagen.flow_from_directory(
        directory=IMAGE_DIR,
        target_size=target_size,
        batch_size=batch_size,
        shuffle=shuffle,
        class_mode=class_mode,
        subset='training')

    test_generator = img_datagen.flow_from_directory(
        directory=IMAGE_DIR,
        target_size=target_size,
        batch_size=batch_size,
        shuffle=shuffle,
        class_mode=class_mode,
        subset='validation')

    return train_generator, test_generator


def data_scaling(data, task="minmax"):
    """
    Scale Data to a similar range

    Parameters
    ----------
    data : dict
        Dictionary containing features and values for the given features
    task : str
        Specific type of scaling that needs to be performed[Standard Scaling vs. Minmax Scaling]

    Returns
    -------
    scaled_data : dict
        Dictionary with the data scaled appropriately

    """ 
    X = data["X"]
    y = data["y"]

    if task == "minmax" or task == "mm":
        scaler = MinMaxScaler()
    elif task == "standard" or task == "s":
        scaler = StandardScaler()
    else:
        raise NameError("Please provide an appropriate scaling option['s' for StandardScaler or 'mm' for MinMax]")

    scaler.fit_transform(X)

    scaled_data = {
        "X": X,
        "y": y
    }
    return scaled_data

def join_arrays(a, b):
    """
    Joining Arrays Row-wise

    Parameters
    ----------
    a : array
        One of the arrays
    b : array
        Second of the arrays
    
    Returns
    -------
    arr : array
        Joined two arrays row wise

    """
    return (np.r_[a, b])



import tensorflow
from tensorflow.keras.models import Sequential
from tensorflow.keras.preprocessing.image import ImageDataGenerator
from tensorflow.keras.layers import Dense, Dropout, Activation, Flatten, BatchNormalization, Conv2D, MaxPooling2D, Reshape, LeakyReLU, Convolution2D, UpSampling2D
from tensorflow.keras.layers import Input
from tensorflow.keras.models import Model, Sequential
from tensorflow.keras.utils import to_categorical
from sklearn.model_selection import train_test_split
import sys
import random
import numpy as np
import pickle

# NN
class NeuralNetwork:
    """This outlines a very rudimentary neural network"""

    def __init__(self):
        """Initializes some variables"""
        self.model = Sequential()
        self.any_hidden_layers = False

        self.input_shape = 'n/a'

        self.x_train = 'n/a'
        self.y_train = 'n/a'
        self.x_test = 'n/a'
        self.y_test = 'n/a'

    def input_data(self, X='n/a', y='n/a'):
        """
        Input your data here.



        Data is for training purposes only
          Evaluating testing data must be done through the 'evaluate_accuracy' function
        """
        if (X == 'n/a' or y == 'n/a'):
            sys.exit('no data was given')
        else:
            self.x_train = X
            self.y_train = y
            self.input_shape = X[0].shape

    def add_input_layer(self, input_shape='n/a'):
        """
        Adds an input layer to your model.



        If you have already executed the 'input_data' command, the input_shape will be automatically calculated.
          You can override this by manually entering a value for 'input_shape'.
        """
        if (self.input_shape == 'n/a'):
            if (input_shape == 'n/a'):
                sys.exit("define the input shape of your data or use the 'input_data' function")
            else:
                self.input_shape = input_shape
        else:
            if (input_shape == 'n/a'):
                pass
            else:
                self.input_shape = input_shape

    def add_hidden_layer(self, neurons=100, activation_func='relu'):
        """
        Adds a hidden layer to your model



        The number of neurons for each layer defaults to 100
          Override this by passing (neurons = x) where x is the number of neurons you want.

        The activation function for each layer defaults to 'relu'
          You will have access to all other activation functions present in the 'keras' library.
          For example, passing the parameter (activation_func = 'sigmoid') will add a layer
          that uses the 'sigmoid' activation function.

        """
        if (self.any_hidden_layers == False):
            self.model.add(Dense(neurons, activation=activation_func, input_shape=self.input_shape))
            self.any_hidden_layers = True
        else:
            self.model.add(Dense(neurons, activation=activation_func))

    def add_output_layer(self, task='classify', neurons=0, activation_func='n/a'):
        """
        Adds the output layer for a neural network



        You will need to input the number of ouput-neurons needed for your model by
        passing (neurons = x) where x is the number of neurons you need.

        The task for the neural network defaults to 'classify'.
          For regression problems, pass (task = 'regressor') when calling the function

        Activation function for 'classify' is defaulted to 'softmax'. Activation function
        for 'regressor' is defaults to 'relu'.
          These can be changed by manually setting (activation_func = 'relu') when
          calling your function.
        """

        task = task.lower()
        activation_func = activation_func.lower()

        if (task == 'regressor' and activation_func == 'n/a'):
            activation_func = 'relu'
        if (task == 'classify' and activation_func == 'n/a'):
            activation_func = 'softmax'

        if (self.any_hidden_layers == False):
            self.model.add(Dense(neurons, activation=activation_func, input_shape=self.input_shape))
        else:
            self.model.add(Dense(neurons, activation=activation_func))

    def run_model(self, X='n/a', y='n/a', epochs=10, batch_size=64, shuffle=True, optimizer='Adam',
                  loss='categorical_crossentropy', metrics=['accuracy']):
        """
        Compiles and trains the neural network



        The training data can either be passed in through the 'input_data' function or
        directly through this 'runmodel' function.
          There is no need to pass in your data twice.

          If you pass in data through this function, it is automatically assumed to be
          training data. To split the data, see the description for 'input_data'

        The number of training iterations (epochs) is defaulted to 10.
          Override this by entering (epochs = x) where x is the number of epochs you need

        'optimizer', 'loss', and 'metrics' are defaulted to 'Adam', 'categorical_crossentropy',
        and ['accuracy'] respectively.
          These can be altered to other values defined in the 'keras' library
        """

        if (X == 'n/a' or y == 'n/a'):
            if (self.x_train == 'n/a' or self.y_train == 'n/a'):
                sys.exit('missing training data')
        else:
            self.x_train = X
            self.y_train = y
        self.model.compile(
            optimizer=optimizer,
            loss=loss,
            metrics=metrics
        )
        self.model.fit(
            x=self.x_train,
            y=to_categorical(self.y_train),
            epochs=epochs,
            batch_size=batch_size,
            shuffle=shuffle
        )

    def create(self, task='classify', output_neurons=0, epochs=10, X='n/a', y='n/a'):
        """
        Creates a rudimentary neural network that will work for most simple tasks
          There is 1 input layer, 2 hidden layers, and 1 output layer



        The number of neurons in the ouput layer must be entered.
          Exception: number of output_neurons defaults to 1 if you specify (task = 'regressor')
          instead of 'classify'

        You can also input your data through this function so that you can avoid
        using the 'input_data' function
          Data only needs to be entered once

        Number of epochs is defaulted to 10
        """
        self.input_data(X=X, y=y)
        self.add_input_layer(self.input_shape)
        self.add_hidden_layer()
        self.add_hidden_layer()
        task = task.lower()
        if (task == 'regressor' and output_neurons == 0):
            output_neurons = 1
            self.add_output_layer(task='regressor', neurons=output_neurons)
        elif (task == 'classify' and output_neurons == 0):
            sys.exit('specify the number of output neurons for your data')
        self.add_output_layer(task=task, neurons=output_neurons)

    def create_and_run(self, task='classify', output_neurons=0, epochs=10, X='n/a', y='n/a', batch_size=64):
        """
        This function creates and runs a simple neural network



        To use this function, you must:
          1. Pass in your data unless you have already run the 'input_data' function

          2. Define the number of output neurons in the network
              You can skip this step if you set (task = 'regressor')

        """
        self.create(task=task, output_neurons=output_neurons, epochs=epochs, X=X, y=y)
        self.run_model(epochs=epochs, batch_size=batch_size)

    def evaluate_accuracy(self, X='n/a', y='n/a', display_accuracy=True):
        """
        Evaluates the performance of your model using testing data based on accuracy



        Testing accuracy is displayed by default
          You can turn this off by setting (display_accuracy = False)
        """
        if (X == 'n/a' or y == 'n/a'):
            if (self.x_test == 'n/a' or self.y_test == 'n/a'):
                print("Missing testing data")
        else:
            self.x_test = X
            self.y_test = y

        self.test_loss, self.test_accuracy = self.model.evaluate(self.x_test, to_categorical(self.y_test))

        if (display_accuracy == True):
            print("Testing accuracy: ", self.test_accuracy)

        return self.test_accuracy

    def predict(self, X='n/a'):
        """
        Makes a prediction for some data



        The network must have been trained by now
        """
        if (X == 'n/a'):
            sys.exit("no data was given")

        return self.model.predict(X)

    def save_model(self, path='n/a'):
        """
        Saves to the neural network to 'path'
          Can be loaded into a new model through the 'load_model' function
        """
        if (path == 'n/a'):
            sys.exit('no path was given')
        self.model.save(path)
        print("model was saved to: ", path)

    def load_model(path):
        """
        Loads a model from a given path

        The following code shows how to load a model into
        the variable 'loaded_model':
          loaded_model = load_model('path')

        """
        return (tensorflow.keras.models.load_model(path))


# CNN
class CNN:
    def __init__(self, num_class, train_data=None, train_labels=None, test_data=None, test_labels=None, type="custom",
                 model=None, from_generator=False):

        if from_generator:
            self.train_data = train_data
            self.test_data = test_data
            self.num_class = num_class
            self.STEPS_PER_EPOCH = train_data.n // train_data.batch_size
            self.VALIDATION_STEPS = test_data.n // test_data.batch_size
            self.from_generator = True
            self.input_shape = train_data.image_shape

        else:
            self.train_data = train_data
            self.train_labels = train_labels
            self.test_data = test_data
            self.test_labels = test_labels
            self.num_class = num_class
            self.from_generator = False
            self.input_shape = train_data.shape[1:]

        self.type = type

        if type == "resnet50":
            base_model = tensorflow.keras.applications.resnet50.ResNet50(input_shape=self.input_shape, weights="imagenet",
                                                              include_top=False)
            x = tensorflow.keras.layers.GlobalAveragePooling2D()(base_model.output)
            output = None
            if num_class > 2:
                output = Dense(num_class, activation='softmax')(x)
            else:
                output = Dense(1, activation='sigmoid')(x)
            self.model = tensorflow.keras.models.Model(inputs=base_model.input, outputs=output)
        
        elif type == "resnet101":
            base_model = tensorflow.keras.applications.ResNet101(input_shape=self.input_shape, weights="imagenet",
                                                              include_top=False)
            x = tensorflow.keras.layers.Flatten()(base_model.output)
            output = None
            if num_class > 2:
                output = Dense(num_class, activation='softmax')(x)
            else:
                output = Dense(1, activation='sigmoid')(x)
            self.model = tensorflow.keras.models.Model(inputs=base_model.input, outputs=output)
        

        elif type == "inception_resnet":
            base_model = tensorflow.keras.applications.InceptionResNetV2(input_shape=self.input_shape, weights="imagenet",
                                                              include_top=False)
            x = tensorflow.keras.layers.Flatten()(base_model.output)
            output = None
            if num_class > 2:
                output = Dense(num_class, activation='softmax')(x)
            else:
                output = Dense(1, activation='sigmoid')(x)
            self.model = tensorflow.keras.models.Model(inputs=base_model.input, outputs=output)
        
        elif type == "inception":
            base_model = tensorflow.keras.applications.InceptionV3(input_shape=self.input_shape, weights="imagenet",
                                                              include_top=False)
            x = tensorflow.keras.layers.Flatten()(base_model.output)
            output = None
            if num_class > 2:
                output = Dense(num_class, activation='softmax')(x)
            else:
                output = Dense(1, activation='sigmoid')(x)
            self.model = tensorflow.keras.models.Model(inputs=base_model.input, outputs=output)

        elif type == "densenet121":
            base_model = tensorflow.keras.applications.DenseNet121(input_shape=self.input_shape, weights="imagenet",
                                                              include_top=False)
            x = tensorflow.keras.layers.Flatten()(base_model.output)
            output = None
            if num_class > 2:
                output = Dense(num_class, activation='softmax')(x)
            else:
                output = Dense(1, activation='sigmoid')(x)
            self.model = tensorflow.keras.models.Model(inputs=base_model.input, outputs=output)
        elif type == "vgg":

            base_model = tensorflow.keras.applications.VGG19(input_shape=self.input_shape, weights="imagenet",
                                                  include_top=False)

            for layer in base_model.layers:
                layer.trainable = False

            x = Flatten()(base_model.output)
            output = None

            if num_class > 2:
                output = Dense(num_class, activation='softmax')(x)
            else:
                output = Dense(1, activation='sigmoid')(x)

            self.model = tensorflow.keras.models.Model(inputs=base_model.input, outputs=output)
        else:
            self.model = Sequential()
            self.model.add(Conv2D(32, (2, 2), padding='same', input_shape=train_data.shape[1:]))
            self.model.add(Activation('relu'))

    def addLayer(self, type, out_activation='softmax', conv_filter=64):

        if self.type != 'custom':
            return

        if type == 'conv' or type == 'convolution':

            self.model.add(Conv2D(conv_filter, (5, 5)))
            self.model.add(Activation('relu'))

        elif type == 'pool' or type == 'pooling':

            self.model.add(MaxPooling2D(pool_size=(2, 2), padding='same'))

        elif type == 'fc' or type == 'fully connected':

            self.model.add(Flatten())
            self.model.add(Dense(512))
            self.model.add(Activation('relu'))

        elif type == 'out' or type == 'output':

            self.model.add(Dense(self.num_class))
            self.model.add(Activation(out_activation))

    def compute(self, loss, train=True, optimizers="adam", lr=0.001, batch=32, num_epochs=10):
        if train:
            if optimizers == "adam":
                self.model.compile(loss=loss, optimizer=tensorflow.keras.optimizers.Adam(learning_rate=lr), metrics=['accuracy'])

            elif optimizers == "sgd":
                self.model.compile(loss=loss, optimizer=tensorflow.keras.optimizers.SGD(learning_rate=lr), metrics=['accuracy'])

            if self.from_generator:
                self.model.fit(self.train_data, validation_data=self.test_data,
                               steps_per_epoch=self.STEPS_PER_EPOCH, validation_steps=self.VALIDATION_STEPS,
                               epochs=num_epochs)
            else:
                self.model.fit(self.train_data, self.train_labels, batch_size=batch, epochs=num_epochs)
        if self.from_generator:
            return self.model.evaluate_generator(self.test_data, verbose=0)
        else:
            return self.model.evaluate(self.test_data, self.test_labels, verbose=0)

    def predict(self, test_data):
        return self.model.predict(test_data)

    def save(self, name="model"):
        self.model.save(f"{name}.h5")

    def load(self, path):
        model = tensorflow.keras.models.load_model(path)


# GAN
class GAN:
    def __init__(self, train_data, inDim):

        if len(train_data.shape) == 3:
            train_data = np.reshape(train_data, (train_data.shape[0], train_data.shape[1], train_data.shape[2], 1))

        # input dimension for generator
        self.inDim = inDim
        self.train_data = train_data

        # number of pixels for completley flat image
        total = 1

        for i in range(1, len(train_data.shape)):
            total *= train_data.shape[i]

        # first layer for generator
        self.gen = Sequential()
        self.gen.add(Dense(256, input_dim=self.inDim, kernel_initializer=tensorflow.keras.initializers.RandomNormal(stddev=0.02)))

        # first layer for disc
        self.disc = Sequential()
        self.disc.add(Dense(1024, input_dim=total, kernel_initializer=tensorflow.keras.initializers.RandomNormal(stddev=0.02)))

    def adddenselayer(self, network, num_neurons=512, activation='leakyrelu'):

        if activation == 'leakyrelu':
            activation = LeakyReLU(0.2)
        else:
            activation = Activation(activation)

        if network == 'g' or network == 'gen' or network == 'generator':
            self.gen.add(Dense(num_neurons, activation=activation))
        elif network == 'd' or network == 'disc' or network == 'discriminator':
            self.disc.add(Dense(num_neurons, activation=activation))

    def adddropout(self, network, rate=0.7):
        if network == 'g' or network == 'gen' or network == 'generator':
            self.gen.add(Dropout(rate))
        elif network == 'd' or network == 'disc' or network == 'discriminator':
            self.disc.add(Dropout(rate))

    def train(self, optimizer='adam', epochs=10, batch_size=128, lr=0.001):
        if optimizer == 'adam':
            optimizer = tensorflow.keras.optimizers.Adam(lr=lr, beta_1=0.5)
        elif optimizer == 'sgd':
            optimizer = tensorflow.keras.optimizers.SGD(learning_rate=lr)

        self.gen.compile(loss='binary_crossentropy', optimizer=optimizer)
        self.disc.compile(loss='binary_crossentropy', optimizer=optimizer)
        print(self.gen.summary())
        self.disc.trainable = False

        gan = Sequential()
        gan.add(self.gen)
        gan.add(self.disc)
        gan.compile(loss='binary_crossentropy', optimizer=optimizer)
        print(gan.summary())
        for i in range(epochs):
            # input for gen
            noise = np.random.randn(batch_size, self.inDim)
            # random images from gen
            genImg = self.gen.predict(noise)

            # taking a random batch of images from training data and storing in rbatch
            s1 = self.train_data.shape[0]
            idx = np.arange(s1)
            np.random.shuffle(idx)
            rBatch = self.train_data[idx[:batch_size], :]

            rBatch = np.reshape(rBatch, (
            batch_size, self.train_data.shape[1] * self.train_data.shape[2] * self.train_data.shape[3]))

            # input is a combination of generated images and random sample
            input = np.concatenate([rBatch, genImg])
            # labels are consists of 1 for random sample and 0 for generated images
            labels = np.zeros(batch_size * 2)
            labels[:batch_size] = 1

            self.disc.trainable = True
            self.disc.train_on_batch(input, labels)

            noise = np.random.randn(batch_size, self.inDim)
            labels = np.ones(batch_size)
            self.disc.trainable = False
            gan.train_on_batch(noise, labels)

    def addconvlayer(self, network, conv_filter=64, filter_size=(3, 3), stride=(1, 1)):
        if network == 'g' or network == 'gen' or network == 'generator':
            self.gen.add(Convolution2D(conv_filter, filter_size=filter_size, strides=stride))
        elif network == 'd' or network == 'disc' or network == 'discriminator':
            self.disc.add(Convolution2D(conv_filter, filter_size=filter_size, strides=stride))

    def addupsampling(self, size=(2, 2)):
        self.gen.add(tensorflow.keras.layers.UpSampling2D(size=size))

    def generate(self, num_img=1):
        noise = np.random.randn(num_img, self.inDim)
        genImg = self.gen.predict(noise)
        return genImg